@@ -179,7 +179,8 @@ class Policy:
                  subject_ou_regexes=None, subject_st_regexes=None, subject_l_regexes=None, subject_c_regexes=None,
                  san_regexes=None, key_types=None, key_reuse=None, cert_authority=None, cert_authority_account_id=None,
                  cert_authority_product_option_id=None, priority=None, modification_date=None, status=None, reason=None,
-                 validity_period=None, recommended_settings=None):
+                 validity_period=None, recommended_settings=None, csr_upload_allowed=None,
+                 key_generated_by_venafi_allowed=None):
         """
         :param str policy_id:
         :param str company_id:
@@ -204,6 +205,8 @@ class Policy:
         :param str reason:
         :param str validity_period:
         :param vaas_utils.RecommendedSettings recommended_settings:
+        :param bool csr_upload_allowed:
+        :param bool key_generated_by_venafi_allowed:
         """
         self.id = policy_id
         self.company_id = company_id
@@ -229,6 +232,11 @@ class Policy:
         self.reason = reason
         self.validity_period = validity_period
         self.recommended_settings = recommended_settings
+        self.csr_upload_allowed = csr_upload_allowed
+        self.key_generated_by_venafi_allowed = key_generated_by_venafi_allowed
+        self.email_regexes = None  # type: list[str]
+        self.ip_constraints_regexes = None  # type: list[str]
+        self.uri_regexes = None  # type: list[str]
 
     def __repr__(self):
         return "Policy:\n" + "\n".join([f"  {k}: {v}" for k, v in (
@@ -268,7 +276,8 @@ class CertificateRequest:
                  csr_origin=CSR_ORIGIN_LOCAL,
                  include_private_key=False,
                  validity_hours=None,
-                 issuer_hint=IssuerHint.DEFAULT):
+                 issuer_hint=IssuerHint.DEFAULT,
+                 use_legacy_pem=False):
         """
         :param str cert_id: Certificate request id. Generating by server.
         :param list[str] san_dns: Alternative names for SNI.
@@ -296,6 +305,7 @@ class CertificateRequest:
         :param bool include_private_key: Indicates if the private key should be returned by the server or not.
         :param int validity_hours: time in hours before the certificate expires.
         :param IssuerHint issuer_hint: Issuer of the certificate. Ignored when platform is not TPP.
+        :param bool use_legacy_pem: Flag that indicates the private key must be in PKCS1 format. Default is PKCS8.
         """
 
         self.chain_option = CHAIN_OPTION_LAST  # "last"
@@ -332,6 +342,7 @@ class CertificateRequest:
         self.include_private_key = include_private_key
         self.validity_hours = validity_hours
         self.issuer_hint = issuer_hint
+        self.use_legacy_pem = use_legacy_pem
 
     def __setattr__(self, key, value):
         if key == "key_password":
@@ -493,9 +504,14 @@ class CertificateRequest:
         else:
             encryption = serialization.NoEncryption()
 
+        if self.use_legacy_pem:
+            pk_format = serialization.PrivateFormat.TraditionalOpenSSL
+        else:
+            pk_format = serialization.PrivateFormat.PKCS8
+
         return self.private_key.private_bytes(
             encoding=serialization.Encoding.PEM,
-            format=serialization.PrivateFormat.TraditionalOpenSSL,
+            format=pk_format,
             encryption_algorithm=encryption,
         ).decode()
 
