@@ -34,8 +34,9 @@ from .policy import PolicySpecification
 from .policy.pm_cloud import (build_policy_spec, validate_policy_spec, AccountDetails, build_cit_request, build_user,
                               UserDetails, build_company, build_apikey, build_app_update_request, get_ca_info,
                               CertificateAuthorityDetails, CertificateAuthorityInfo, build_account_details,
-                              build_app_create_request)
-from .vaas_utils import AppDetails, RecommendedSettings, EdgeEncryptionKey, zip_to_pem, value_matches_regex
+                              build_app_create_request, build_team, get_cit_data_from_response, build_owner_json,
+                              supported_rsa_key_sizes, supported_elliptic_curves)
+from .vaas_utils import AppDetails, OwnerIdsAndTypes, RecommendedSettings, EdgeEncryptionKey, zip_to_pem, value_matches_regex
 
 TOKEN_HEADER_NAME = "tppl-api-key"  # nosec
 APPLICATION_SERVER_TYPE_ID = "784938d1-ef0d-11eb-9461-7bb533ba575b"
@@ -50,6 +51,15 @@ CSR_ATTR_PROVINCE = 'state'
 CSR_ATTR_COUNTRY = 'country'
 CSR_ATTR_SANS_BY_TYPE = 'subjectAlternativeNamesByType'
 CSR_ATTR_SANS_DNS = 'dnsNames'
+CSR_ATTR_SANS_IP_ADDR = 'ipAddresses'
+CSR_ATTR_SANS_EMAIL_ADDR = 'rfc822Names'
+CSR_ATTR_SANS_URIS = 'uniformResourceIdentifiers'
+CSR_ATTR_KEY_TYPE_PARAMS = 'keyTypeParameters'
+CSR_ATTR_KEY_TYPE = 'keyType'
+CSR_ATTR_KEY_LENGTH = 'keyLength'
+CSR_ATTR_KEY_CURVE = 'keyCurve'
+OWNER_TYPE_USER = "USER"
+OWNER_TYPE_TEAM = "TEAM"
 
 log = get_child("connection-vaas")
 
@@ -87,6 +97,9 @@ class URLS:
     ISSUING_TEMPLATES_UPDATE = ISSUING_TEMPLATES + "/{}"
     USER_ACCOUNTS = API_VERSION + "useraccounts"
     DEK_PUBLIC_KEY = API_VERSION + "edgeencryptionkeys/{}"
+    USERS_USERNAME = API_VERSION + "users/username/{}"
+    USERS_ID = API_VERSION + "users/{}"
+    TEAMS_ID = API_VERSION + "teams"
 
 
 class CondorChainOptions:
@@ -119,6 +132,16 @@ def _parse_zone(zone):
     return app_name, cit_alias
 
 
+def create_owner(owner_type, owner_id):
+    owner = OwnerIdsAndTypes(owner_type, owner_id)
+    return owner
+
+
+def resolve_apikey_owner(user_details):
+    owner = create_owner(OWNER_TYPE_USER, user_details.user.user_id)
+    return owner
+
+
 class CloudConnection(CommonConnection):
     def __init__(self, token, url=None, http_request_kwargs=None):
         super().__init__()
@@ -247,13 +270,23 @@ class CloudConnection(CommonConnection):
             d['status'] if 'status' in d else None,
             d['reason'] if 'reason' in d else None,
             d['validityPeriod'] if 'validityPeriod' in d else None,
-            None
+            None,
+            d['csrUploadAllowed'] if 'csrUploadAllowed' in d else None,
+            d['keyGeneratedByVenafiAllowed'] if 'keyGeneratedByVenafiAllowed' in d else None
         )
+        policy.email_regexes = d['sanRfc822NameRegexes'] if 'sanRfc822NameRegexes' in d else None
+        policy.ip_constraints_regexes = d['sanIpAddressRegexes'] if 'sanIpAddressRegexes' in d else None
+        policy.uri_regexes = d['sanUniformResourceIdentifierRegexes'] if 'sanUniformResourceIdentifierRegexes' in d \
+            else None
+
         for kt in d.get('keyTypes', []):
             key_type = kt['keyType'].lower()
             if key_type == KeyType.RSA:
                 for s in kt['keyLengths']:
                     policy.key_types.append(KeyType(key_type, s))
+            elif key_type == KeyType.ECDSA:
+                for s in kt["keyCurves"]:
+                    policy.key_types.append(KeyType(key_type, s))
             else:
                 log.error(f"Unknown key type: {kt['keyType']}")
                 raise ServerUnexptedBehavior
@@ -278,7 +311,11 @@ class CloudConnection(CommonConnection):
                 rs['keyReuse'] if 'keyReuse' in rs else None
             )
             if 'key' in rs:
-                kt = KeyType(rs['key']['type'], rs['key']['length'])
+                key = rs['key']
+                k_type = key['type']
+                kl = key['length'] if 'length' in key else None
+                kc = key['curve'] if 'curve' in key else None
+                kt = KeyType(k_type, kl or kc)
                 settings.keyType = kt
 
             return settings
@@ -607,22 +644,118 @@ class CloudConnection(CommonConnection):
             raise VenafiError('User Details not found')
 
         app_details = self._get_app_details_by_name(app_name)
+        for_update, owners_list = self.resolve_owners(policy_spec.users, user_details)
         if app_details:
-            # Application exists. Update with cit
-            if not self._policy_exists(zone):
+            # Application exists. Update with cit and owners
+            if for_update and len(policy_spec.users) == 0:
+                log.debug("No users provided in the policy specification")
+            else:
+                owner_list = build_owner_json(owners_list)
+                app_details.owner_ids_and_types = owner_list
+            if app_details.cit_alias_id_map:
                 # Only link cit with Application when cit is not already associated with Application
-                app_req = build_app_update_request(app_details, resp_cit_data)
-                status, data = self._put(URLS.APP_BY_ID.format(app_details.app_id), app_req)
-                if status != HTTPStatus.OK:
-                    raise VenafiError(f"Could not update Application [{app_name}] with cit [{resp_cit_data}]")
+                cit_map = app_details.cit_alias_id_map
+                cit_id, cit_name = get_cit_data_from_response(cit_data)
+                cit_map[cit_name] = cit_id
+            app_req = build_app_update_request(app_details, cit_map)
+            status, data = self._put(URLS.APP_BY_ID.format(app_details.app_id), app_req)
+            if status != HTTPStatus.OK:
+                raise VenafiError(f"Could not update Application [{app_name}] with cit [{resp_cit_data}]")
         else:
             # Application does not exist. Create one
-            app_req = build_app_create_request(app_name, user_details, resp_cit_data)
+            app_req = build_app_create_request(app_name, owners_list, resp_cit_data)
             status, data = self._post(URLS.APPLICATIONS, app_req)
             if status != HTTPStatus.CREATED:
                 raise VenafiError(f"Could not create application [{app_name}]")
         return
 
+    def resolve_owners(self, users_list, user_details):
+        owners_list = list()
+        for_update = False
+        if not users_list:
+            # When no users are provided on the list, adds apikey user as owner
+            api_owner = resolve_apikey_owner(user_details)
+            owners_list.append(api_owner)
+            for_update = True
+        else:
+            # Resolving the usernames list
+            # Creating a higher level Teams object to cache the response
+            teams_list = list()
+            users_list = list(set(users_list))
+            for username in users_list:
+                cloud_owner = self.resolve_user_to_cloud_owner(username)
+                if cloud_owner:
+                    owners_list.append(cloud_owner)
+                else:
+                    team_owner = self.resolve_user_to_cloud_team(username)
+                    if not team_owner:
+                        raise VenafiError(f"Unable to find identity [{username}]")
+                    owners_list.append(team_owner)
+        return for_update, owners_list
+
+    def resolve_user_to_cloud_owner(self, username):
+        user = self.get_vaas_identity(username)
+        if not user:
+            return None
+        owner = create_owner(OWNER_TYPE_USER, user.user_id)
+        return owner
+
+    def resolve_user_to_cloud_team(self, username):
+        teams = list()
+        data_teams = None
+        try:
+            status, response = self._get(URLS.TEAMS_ID)
+            data_teams = response['teams']
+        except VenafiError as err:
+            log.debug(f"Error while getting team [{username}]: {err.args[0]}")
+        # date_teams will never be empty, we are always expecting at least one team
+        for data in data_teams:
+            team = build_team(data)
+            teams.append(team)
+        if len(teams) > 0:
+            for team in teams:
+                if team.name == username:
+                    owner = create_owner(OWNER_TYPE_TEAM, team.team_id)
+                    return owner
+        return None
+
+    def get_vaas_identity(self, username):
+        if not username or username == "":
+            raise VenafiError("Username cannot be empty")
+        url = URLS.USERS_USERNAME.format(username)
+        try:
+            status, response = self._get(url)
+            if status == HTTPStatus.NOT_FOUND:
+                return None
+            identities = response['users'][0]
+            identity = build_user(identities)
+            return identity
+        except VenafiError as err:
+            log.debug(f"Unable to find identity [{username}] of type USER: {err.args[0]}")
+
+    def resolve_cloud_owners_names(self, zone):
+        app_name, cit_alias = _parse_zone(zone)
+        app_details = self._get_app_details_by_name(app_name)
+        users_list = list()
+        teams_response = list()
+        for owner in app_details.owner_ids_and_types:
+            if owner['ownerType'] == OWNER_TYPE_USER:
+                status, data = self._get(URLS.USERS_ID.format(owner['ownerId']))
+                user = build_user(data)
+                users_list.append(user.username)
+            elif owner['ownerType'] == OWNER_TYPE_TEAM:
+                if not teams_response:
+                    status, data = self._get(URLS.TEAMS_ID)
+                    data_team = data['teams']
+                    for t in data_team:
+                        team = build_team(t)
+                        teams_response.append(team)
+                if teams_response:
+                    for team in teams_response:
+                        if team.team_id == owner['ownerId']:
+                            users_list.append(team.name)
+        return users_list
+
     def _get_ca_details(self, ca_name):
         """
         :param str ca_name:
@@ -730,7 +863,7 @@ class CloudConnection(CommonConnection):
         if request.organization:
             if ps.policy and ps.policy.subject:
                 policy_orgs = ps.policy.subject.orgs
-                valid = value_matches_regex(value=request.organization,pattern_list=policy_orgs)
+                valid = value_matches_regex(value=request.organization, pattern_list=policy_orgs)
                 if not valid:
                     org_str = "Organization"
                     log.error(MSG_VALUE_NOT_MATCH_POLICY.format(org_str, f"{org_str}s", request.organization,
@@ -756,7 +889,7 @@ class CloudConnection(CommonConnection):
                     log.error(MSG_VALUE_NOT_MATCH_POLICY.format(ou_str, f"{ou_str}s", request.organizational_unit,
                                                                 policy_ous))
                     raise ClientBadData
-            csr_attr_map[CSR_ATTR_ORG_UNIT] = request.organizational_unit
+            csr_attr_map[CSR_ATTR_ORG_UNIT] = org_units
         elif ps.defaults and ps.defaults.subject and ps.defaults.subject.org_units:
             csr_attr_map[CSR_ATTR_ORG_UNIT] = ps.defaults.subject.org_units
 
@@ -775,7 +908,7 @@ class CloudConnection(CommonConnection):
 
         if request.province:
             if ps.policy and ps.policy.subject:
-                policy_provinces = ps.policy.subject.localities
+                policy_provinces = ps.policy.subject.states
                 valid = value_matches_regex(value=request.province, pattern_list=policy_provinces)
                 if not valid:
                     province_str = "Province"
@@ -800,12 +933,63 @@ class CloudConnection(CommonConnection):
             csr_attr_map[CSR_ATTR_COUNTRY] = ps.defaults.subject.country
 
         if len(request.san_dns) > 0:
-            sans = {
-                CSR_ATTR_SANS_DNS: request.san_dns
-                # TODO: Other sans should be added here
-            }
+            sans = dict()
+            if request.san_dns and len(request.san_dns) > 0:
+                sans[CSR_ATTR_SANS_DNS] = request.san_dns
+            if request.ip_addresses and len(request.ip_addresses) > 0:
+                sans[CSR_ATTR_SANS_IP_ADDR] = request.ip_addresses
+            if request.email_addresses and len(request.email_addresses) > 0:
+                sans[CSR_ATTR_SANS_EMAIL_ADDR] = request.email_addresses
+            if request.uniform_resource_identifiers and len(request.uniform_resource_identifiers) > 0:
+                sans[CSR_ATTR_SANS_URIS] = request.uniform_resource_identifiers
+
             csr_attr_map[CSR_ATTR_SANS_BY_TYPE] = sans
 
+        if request.key_type:
+            if ps.policy and ps.policy.key_pair:
+                policy_kts = ps.policy.key_pair.key_types
+                valid = value_matches_regex(value=request.key_type.key_type.upper(), pattern_list=policy_kts)
+                if not valid:
+                    kt_str = "Key Type"
+                    log.error(MSG_VALUE_NOT_MATCH_POLICY.format(kt_str, f"{kt_str}s", request.key_type.key_type,
+                                                                policy_kts))
+                    raise ClientBadData
+                req_kt_option = request.key_type.option
+                if request.key_type.key_type.lower() == KeyType.RSA:
+                    policy_rsa_sizes = ps.policy.key_pair.rsa_key_sizes
+                    valid = True if req_kt_option in policy_rsa_sizes else False
+                    if not valid:
+                        rsa_str = "RSA Key Size"
+                        log.error(MSG_VALUE_NOT_MATCH_POLICY.format(rsa_str, f"{rsa_str}s", req_kt_option,
+                                                                    policy_rsa_sizes))
+                        raise ClientBadData
+                elif request.key_type.key_type.lower() == KeyType.ECDSA:
+                    policy_ecs = ps.policy.key_pair.elliptic_curves
+                    valid = value_matches_regex(value=req_kt_option, pattern_list=policy_ecs)
+                    if not valid:
+                        ec_str = "Elliptic Curve"
+                        log.error(MSG_VALUE_NOT_MATCH_POLICY.format(ec_str, f"{ec_str}s", req_kt_option, policy_ecs))
+                        raise ClientBadData
+
+            kt_param = {
+                CSR_ATTR_KEY_TYPE: request.key_type.key_type.upper()
+            }
+            kt_type = request.key_type.key_type.lower()
+            if kt_type == KeyType.RSA:
+                kt_param[CSR_ATTR_KEY_LENGTH] = request.key_type.option
+            elif kt_type == KeyType.ECDSA:
+                kt_param[CSR_ATTR_KEY_CURVE] = request.key_type.option.upper()
+
+            csr_attr_map[CSR_ATTR_KEY_TYPE_PARAMS] = kt_param
+        elif ps.defaults and ps.defaults.key_pair:
+            kt_param = {
+                CSR_ATTR_KEY_TYPE: ps.defaults.key_pair.key_type.upper()
+            }
+            if ps.defaults.key_pair.key_type.lower() == KeyType.RSA:
+                kt_param[CSR_ATTR_KEY_LENGTH] = ps.defaults.key_pair.rsa_key_size
+            elif ps.defaults.key_pair.key_type.lower() == KeyType.ECDSA:
+                kt_param[CSR_ATTR_KEY_CURVE] = ps.defaults.key_pair.elliptic_curve
+
         return csr_attr_map
 
     def _get_policy(self, zone, subject_cn_to_str):
@@ -825,6 +1009,8 @@ class CloudConnection(CommonConnection):
             raise VenafiError("Certificate Authority info not found")
 
         ps = build_policy_spec(cit, info, subject_cn_to_str)
+        users_list = self.resolve_cloud_owners_names(zone)
+        ps.users = users_list
         return ps
 
     def _get_dek_hash(self, cert_id):
@@ -875,3 +1061,4 @@ class CloudConnection(CommonConnection):
 
         cert, chain, private_key = zip_to_pem(data, request.chain_option)
         return Certificate(cert=cert, chain=chain, key=private_key)
+
