@@ -23,6 +23,7 @@ from cryptography import x509
 from cryptography.hazmat.backends import default_backend
 from cryptography.x509 import SignatureAlgorithmOID as AlgOID
 from six.moves.urllib import parse as url_parse
+from types import SimpleNamespace
 
 from .common import CertField, CommonConnection, CertificateRequest, CSR_ORIGIN_LOCAL, CSR_ORIGIN_PROVIDED, \
     CSR_ORIGIN_SERVICE, KeyType, CHAIN_OPTION_LAST, CHAIN_OPTION_FIRST, CHAIN_OPTION_IGNORE, Policy, ZoneConfig
@@ -32,7 +33,7 @@ from .http_status import HTTPStatus
 from .pem import parse_pem
 from .policy import RPA, SPA
 from .policy.pm_tpp import TPPPolicy, is_service_generated_csr, SetAttrResponse, validate_policy_spec, \
-    get_int_value
+    get_int_value, build_identity_entry
 from .ssh_utils import SSHCertRequest, SSHCertResponse, build_tpp_retrieve_request, SSHResponse, \
     SSHRetrieveResponse, build_tpp_request, SSHCATemplateRequest, SSHConfig, PATH_SEPARATOR, CA_ROOT_PATH, \
     SSHTPPCADetails
@@ -66,6 +67,8 @@ class URLS:
     POLICY_CREATE = API_BASE_URL + "config/create"
     POLICY_SET_ATTRIBUTE = API_BASE_URL + "config/writepolicy"
     POLICY_CLEAR_ATTRIBUTE = API_BASE_URL + "config/clearpolicyattribute"
+    POLICY_BROWSE_IDENTITY = API_BASE_URL + "identity/browse"
+    POLICY_VALIDATE_IDENTITY = API_BASE_URL + "identity/validate"
 
     SSH_BASE_URL = API_BASE_URL + "SSHCertificates/"
     SSH_CERTIFICATE_REQUEST = SSH_BASE_URL + "request"
@@ -86,6 +89,10 @@ class AbstractTPPConnection(CommonConnection):
     ARG_CHECK_TOKEN = 'check_token'  # nosec
     ARG_INCLUDE_TOKEN_HEADER = 'include_token_header'  # nosec
     ARG_DATA = 'data'
+    IDENTITY_USER = 1
+    IDENTITY_SECURITY_GROUP = 2
+    IDENTITY_DISTRIBUTION_GROUP = 8
+    ALL_IDENTITIES = IDENTITY_USER + IDENTITY_SECURITY_GROUP + IDENTITY_DISTRIBUTION_GROUP
 
     def auth(self):
         raise NotImplementedError
@@ -175,9 +182,12 @@ class AbstractTPPConnection(CommonConnection):
         log.debug(f"Getting certificate status for id {cert_request.id}")
 
         retrieve_request = dict(CertificateDN=cert_request.id,
-                                Format="base64",
+                                Format="Base64 (PKCS #8)",
                                 IncludeChain=True)
 
+        if cert_request.use_legacy_pem:
+            retrieve_request["Format"] = "base64"
+
         if cert_request.csr_origin == CSR_ORIGIN_SERVICE:
             retrieve_request['IncludePrivateKey'] = cert_request.include_private_key
             if cert_request.key_password:
@@ -470,7 +480,7 @@ class AbstractTPPConnection(CommonConnection):
 
         log.info("Building Policy Specification")
         spec = tpp_policy.to_policy_spec()
-
+        spec.users = self.retrieve_usernames_from_tpp_contacts(policy_name)
         return spec
 
     def set_policy(self, zone, policy_spec):
@@ -483,6 +493,8 @@ class AbstractTPPConnection(CommonConnection):
         tpp_policy = TPPPolicy.build_tpp_policy(policy_spec)
         name = self._normalize_zone(zone)
         tpp_policy.name = name
+        identities_list = self.resolve_tpp_contacts(tpp_policy.contact)
+        tpp_policy.contact = identities_list
 
         create_policy = False
         policy_exists = self._policy_exists(name)
@@ -563,6 +575,43 @@ class AbstractTPPConnection(CommonConnection):
 
         return
 
+    def resolve_tpp_contacts(self, contacts):
+        identities_id_list = list()
+        if contacts:
+            contacts = list(set(contacts))
+            for contact in contacts:
+                identity = self.get_tpp_identity(contact)
+                identities_id_list.append(identity.prefixed_universal)
+        return identities_id_list
+
+    def get_tpp_identity(self, username):
+        if not username or username == "":
+            raise VenafiError("Identity cannot be empty")
+        data = {
+            'Filter': username,
+            'Limit': 2,
+            'IdentityType': self.ALL_IDENTITIES
+        }
+        try:
+            status, response = self._post(URLS.POLICY_BROWSE_IDENTITY, data=data)
+        except VenafiError as err:
+            log.debug(f"Error while getting contact [{username}]: {err.args[0]}")
+        if len(response['Identities']) == 0:
+            raise VenafiError(f"The username [{username}] was not found.")
+        identity = None
+        if len(response['Identities']) > 1:
+            for identy_entry in response['Identities']:
+                identy_user = build_identity_entry(identy_entry)
+                if identy_user.name.casefold() == username:
+                    identity = identy_user
+                    break
+        else:
+            identities = response['Identities'][0]
+            identity = build_identity_entry(identities)
+        if not identity:
+            raise VenafiError(f"The username [{username}] was not found.")
+        return identity
+
     def request_ssh_cert(self, request):
         """
 
@@ -952,3 +1001,65 @@ class AbstractTPPConnection(CommonConnection):
                                   f"Message: {response_object.error_msg}")
         else:
             raise ServerUnexptedBehavior(f"Server returns {status} status on requesting SSH CA details")
+
+    def _get_policy_attr(self, zone, attr_name):
+        """
+        :param str zone:
+        :param str attr_name:
+        :rtype: tuple[str, SetAttrResponse]
+        """
+        data = {
+            'ObjectDN': zone,
+            'Class': POLICY_ATTR_CLASS,
+            'AttributeName': attr_name,
+        }
+
+        # TODO: Change _post() with post(args)
+        status, response = self._post(URLS.FIND_POLICY, data=data)
+        if status != HTTPStatus.OK:
+            raise ServerUnexptedBehavior(f"Got status {status} from server")
+        return status, response
+
+    def retrieve_usernames_from_tpp_contacts(self, zone):
+        users_list = list()
+        contacts = list()
+        try:
+            status, response = self._get_policy_attr(zone, SPA.TPP_CONTACT)
+            contacts = response['Values']
+        except VenafiError as err:
+            log.debug(f"Error while getting attribute [Contact] value from policy [{zone}]: {err.args[0]}")
+        contacts = list(set(contacts))
+        for prefixed_universal in contacts:
+            try:
+                identity = self.validate_identity(prefixed_universal)
+                users_list.append(identity.name)
+            except VenafiError as err:
+                log.debug(f"Error while validating contact [{prefixed_universal}]: {err.args[0]}")
+        return users_list
+
+    def browse_identities(self, username):
+        data = {
+            'Filter': username,
+            'Limit': 2,
+            'IdentityType': self.ALL_IDENTITIES
+        }
+        try:
+            status, response = self._post(URLS.POLICY_BROWSE_IDENTITY, data=data)
+        except VenafiError as err:
+            log.debug(f"Error while getting contact [{username}]: {err.args[0]}")
+
+        if response['Identities'] > 1:
+            raise VenafiError("Extraneous information returned in the identity response. "
+                              "Expected size: 1, found: 2\n")
+        identity = build_identity_entry(response)
+        return identity
+
+    def validate_identity(self, prefixed_universal):
+        data = {
+            'ID': {
+                'PrefixedUniversal': prefixed_universal
+            }
+        }
+        status, response = self._post(URLS.POLICY_VALIDATE_IDENTITY, data=data)
+        identity = build_identity_entry(response['ID'])
+        return identity
